import sdtfile
import numpy as np
import zipfile  
    
def read_sdt_info_brukerSDT(filename):  
    """ 
    modified from CGohlke sdtfile.py to read bruker 150 card data
    gives tarr, x.shape,y.shape,t.shape,c.shape
    """
    ## HEADER
    measure_info = []
    dtype = np.dtype(sdtfile.sdtfile.MEASURE_INFO)
    with open(filename, 'rb') as fh:
        ## HEADER
        header = np.rec.fromfile(fh, dtype=sdtfile.sdtfile.FILE_HEADER, shape=1, byteorder='<')
        fh.seek(header.meas_desc_block_offs[0])
        for _ in range(header.no_of_meas_desc_blocks[0]):
            measure_info.append(
                np.rec.fromfile(fh, dtype=dtype, shape=1, byteorder='<'))
            fh.seek(header.meas_desc_block_length[0] - dtype.itemsize, 1)
    
    times = []
    block_headers = []

    try:
        routing_channels_x = measure_info[0]['image_rx'][0]
    except:
        routing_channels_x = 1

    offset = header.data_block_offs[0]
 
    with open(filename, 'rb') as fh:
        for _ in range(header.no_of_data_blocks[0]): ## 
            fh.seek(offset)
            # read data block header
            bh = np.rec.fromfile(fh, dtype=sdtfile.sdtfile.BLOCK_HEADER, shape=1,
                                 byteorder='<')[0]
            block_headers.append(bh)
            # read data block
            mi = measure_info[bh.meas_desc_block_no]
            
            dtype = sdtfile.sdtfile.BlockType(bh.block_type).dtype
            dsize = bh.block_length // dtype.itemsize
            
            t = np.arange(mi.adc_re[0], dtype=np.float64)
            t *= mi.tac_r / float(mi.tac_g * mi.adc_re)
            times.append(t)
            offset = bh.next_block_offs
        return (header.data_block_offs[0], times, [mi.scan_x[0], mi.scan_y[0], mi.adc_re[0], routing_channels_x])
    
    
def read_sdt150(filename):
    """
    filenamme: str
    output: dataSDT: np.ndarray
    read sdt file and return data in CXYT format
    """
    import warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    offset, t, XYTC = read_sdt_info_brukerSDT(filename)
    try: 
        # if the input file is sdt_zipped
        with zipfile.ZipFile(filename) as myzip:
            z1 = myzip.infolist()[0]  # "data_block"
            with myzip.open(z1.filename) as myfile:
                dataspl = myfile.read()
    except:
        # if the input file is unzipped 
        with open(filename, 'rb') as myfile:
            myfile.seek(offset)
            dataspl = myfile.read()
            
    dataSDT = np.fromstring(dataspl, np.uint16)
    
    if XYTC[3] == 1:
        # reduce the 4D data to 3d (CXYT to XYT)
        dataSDT = dataSDT[:XYTC[0] * XYTC[1] * XYTC[2]].reshape([XYTC[0], XYTC[1], XYTC[2]])
    # reshape XYTC to CXYT
    elif XYTC[3] > 1:
        # Check for empty channels and filter them out
        non_empty_channels =  len(dataSDT) // (XYTC[0] * XYTC[1] * XYTC[2])
        if non_empty_channels == 1:
            # If only one channel is present, reshape to 3D
            dataSDT = dataSDT[:XYTC[0] * XYTC[1] * XYTC[2]].reshape([XYTC[0], XYTC[1], XYTC[2]])
        else:
            dataSDT = dataSDT[:XYTC[0] * XYTC[1] * XYTC[2] * XYTC[3]].reshape([XYTC[3], XYTC[0], XYTC[1], XYTC[2]])

   
    return dataSDT


from .csv_maker import generate_csv_with_column_prompt
from .sdt_reading import read_sdt150
from .Decay_feature import SingleCellFeatureExtractor
from .PCA_analysis import SingleCellPCAAnalysis
from .Optimal_Transport import Optimal_Transport

__all__ = [
    "generate_csv_with_column_prompt",
    "read_sdt150",
    "SingleCellFeatureExtractor",
    "SingleCellPCAAnalysis",
    "Optimal_Transport"
]


# import statements
import os
import csv

# # input statements to obtain path to data + name of output csv
# home_dir = os.path.expanduser("~")
# file_path = input("Please input path to folder containing data: ")
# output_csv = input("Please input the name of the resulting csv file (.csv is NOT required): ") + ".csv"
# output_csv = os.path.join(os.getcwd(), output_csv)
# root_dir = file_path

# #  ----- working ----- #

# # test path: /scr/data/skala_redox_data/241004_New_2DG+A549/A549_Cyanide

# # keep track # of '_'
# num_underscore = 0

# # stores list of list of data extracted from each file name
# list_extracted_features = []

# # stores list of all sub-folder names
# file_names = []

# # list that stores all rows of csv file
# write_rows = []

# # stores full path of each sub-folder
# paths = []

# # stores list of headers prior to update
# first_headers = ["file_name", "file_path"]

# # stores list of updated header names
# updated_col_names = ["file_name", "file_path"]



# # extracts features from name
# for entry in os.listdir(root_dir):
# 	full_path = os.path.join(root_dir, entry)
# 	file_names.append(entry)
# 	paths.append(full_path)
	
# 	if os.path.isdir(full_path):
# 		parts = entry.split("_")
# 		num_underscore = len(parts)
# 		list_extracted_features.append(parts)



# # adds features obtained to new row in csv file        
# def add_features():
# 	for i in range(len(list_extracted_features)):
# 		list = []
# 		list.append(file_names[i])
# 		list.append(paths[i])

# 		for j in range(len(list_extracted_features[0])):
# 			list.append(list_extracted_features[i][j])

# 		write_rows.append(list)



# # creates the csv file and outputs first 3 rows of csv file
# def create_csv():
# 	with open(output_csv, mode='w', newline='') as csv_file:
# 		writer = csv.writer(csv_file)
# 		for i in range(num_underscore):
# 			first_headers.append("column_" + str(i))
			
		
# 		writer.writerow(first_headers)

# 		print("\n---------- CSV Preview (header + first 2 rows): ----------\n")
# 		print(", ".join(first_headers))
# 		for row in write_rows[:2]:
# 			print(", ".join(row))
# 		print("\n")

# 		writer.writerows(write_rows)



# # asks user to input column names
# def update_col_names():
# 	# updated list of headers
# 	print("Please input the following names for columns:\n  -cell_line\n  -condition\n  -dish\n  -channel_type\n  -cell_type\n")
# 	possible_inputs = ["cell_line", "condition", "dish", "channel_type", "cell_type"]
# 	for i in range(num_underscore):
# 		new_col_name = input(first_headers[i+2] + ": ")
# 		if new_col_name not in possible_inputs:
# 			raise Exception("INVALID INPUT")
# 		updated_col_names.append(str(new_col_name))
	


# # updates each row
# def update_row():
# 	# index of dish
# 	index_dish = updated_col_names.index("dish") - 2

# 	# index of channel type
# 	index_ct = updated_col_names.index("channel_type") - 2

# 	# changes Dish# -> # in each row
# 	for i in range(len(list_extracted_features)):
# 		change_dish = list_extracted_features[i][index_dish].upper()		
# 		if "DISH" in change_dish:
# 			dish_num = change_dish.replace("DISH", "")
# 			list_extracted_features[i][index_dish] = dish_num

# 		else:
# 			raise Exception("WRONG COLUMN NAME")

# 	# changes #n or #f -> n or f
# 	for i in range(len(list_extracted_features)):
# 		change_ct_list = list(list_extracted_features[i][index_ct])

# 		if 'f' in change_ct_list:
# 			list_extracted_features[i][index_ct] = 'f'

# 		elif 'n' in change_ct_list:
# 			list_extracted_features[i][index_ct] = 'n'
		
# 		else:
# 			raise Exception("INVALID CHANNEL TYPE")

# 	for i in range(len(file_names)):
# 		list_extracted_features[i].insert(0, file_names[i])
# 		list_extracted_features[i].insert(1, paths[i])




# # updates the csv files with newly inputted headers
# def update_csv():
# 	with open(output_csv, mode='w', newline='') as csv_file:
# 		writer = csv.writer(csv_file)
# 		writer.writerow(updated_col_names)
# 		writer.writerows(list_extracted_features)
	



# add_features()
# create_csv()
# update_col_names()
# update_row()
# update_csv()


def generate_csv_with_column_prompt(input_folder, output_csv_name):
	"""
	Wrapper to call the original csv generator with column name prompt.
	Still requires manual input for columns (as originally designed).
	"""
	global file_names, paths, list_extracted_features, updated_col_names
	file_names, paths, list_extracted_features = [], [], []
	updated_col_names = ["file_name", "file_path"]

	# Re-run logic manually within this function as per original flow
	root_dir = input_folder
	output_csv = output_csv_name if output_csv_name.endswith(".csv") else output_csv_name + ".csv"
	output_csv = os.path.join(os.getcwd(), output_csv)

	# Extract
	for entry in os.listdir(root_dir):
			full_path = os.path.join(root_dir, entry)
			file_names.append(entry)
			paths.append(full_path)
			if os.path.isdir(full_path):
					parts = entry.split("_")
					list_extracted_features.append(parts)

	num_underscore = len(list_extracted_features[0])

	# Prompt user for column names
	print("Please input the following names for columns:\n  -cell_line\n  -condition\n  -dish\n  -channel_type\n  -cell_type\n")
	possible_inputs = ["cell_line", "condition", "dish", "channel_type", "cell_type"]
	for i in range(num_underscore):
			new_col_name = input("column_" + str(i) + ": ")
			if new_col_name not in possible_inputs:
					raise Exception("INVALID INPUT")
			updated_col_names.append(str(new_col_name))

	# Fix and insert metadata
	index_dish = updated_col_names.index("dish") - 2
	index_ct = updated_col_names.index("channel_type") - 2

	for i in range(len(list_extracted_features)):
			dish_val = list_extracted_features[i][index_dish].upper()
			if "DISH" in dish_val:
					list_extracted_features[i][index_dish] = dish_val.replace("DISH", "")
			else:
					raise Exception("Expected 'DISH#' format in 'dish' column")

			ct_val = list_extracted_features[i][index_ct]
			if 'f' in ct_val:
					list_extracted_features[i][index_ct] = 'f'
			elif 'n' in ct_val:
					list_extracted_features[i][index_ct] = 'n'
			else:
					raise Exception("Expected channel type to include 'f' or 'n'")
			
			list_extracted_features[i].insert(0, file_names[i])
			list_extracted_features[i].insert(1, paths[i])

	# Write CSV
	with open(output_csv, mode='w', newline='') as csv_file:
			writer = csv.writer(csv_file)
			writer.writerow(updated_col_names)
			writer.writerows(list_extracted_features)
